from typing import Any, overload

from rich.console import Console
from rich.prompt import Confirm, Prompt
from rich.theme import Theme

_THEME = Theme(
    {
        # custom rgb hex colors most likely from D3 category 10
        "debug": "bold grey11",
        "info": "bold cyan",
        "warn": "bold yellow",
        "error": "bold red",
        "path": "underline #1f77b4",  # a blue
        "url": "underline #1f77b4",  # a blue
        "hint": "italic",
        "package": "bold #ff7f0e",  # an orange
        "addon": "bold #2ca02c",  # a green
        "promptq": "bold",  # for prompt questions
        "prompta": "italic #bcbd22",  # a yellow, for prompt feedback
        "key": "#d62728",  # a red, for things pressed on the keyboard
        "command": "#2ca02c",  # a green, for commands that can be run
        "flavor": "#9467bd",  # a purple, for wow flavors
    }
)

_STDERR_CONSOLE = Console(stderr=True, highlight=False, theme=_THEME)
_STDOUT_CONSOLE = Console(stderr=False, highlight=False, theme=_THEME)


def print(text: Any, stderr: bool = True, newline: bool = True) -> None:
    """
    Print text to the console.

    If `stderr` is True, print to stderr. Otherwise, print to stdout.

    If `newline` is True, print a newline character after the string.
    """
    if stderr:
        console = _STDERR_CONSOLE
    else:
        console = _STDOUT_CONSOLE

    console.print(text, end="\n" if newline else "")


def debug(text: str, newline: bool = True) -> None:
    print(f"[debug]\\[DEBUG][/debug] {text}", newline=newline)


def info(text: str, newline: bool = True) -> None:
    print(f"[info]\\[INFO][/info] {text}", newline=newline)


def warn(text: str, newline: bool = True) -> None:
    print(f"[warn]\\[WARN][/warn] {text}", newline=newline)


def error(text: str, newline: bool = True) -> None:
    print(f"[error]\\[ERROR][/error] {text}", newline=newline)


def _render_as_style(text: str, style: str) -> str:
    """Produce a text in the style."""
    return _THEME.styles[style].render(text)


# this is only overloaded because rich overloads it. unnecessary...
@overload
def prompt_ask(prompt: str) -> str:
    ...


@overload
def prompt_ask(prompt: str, default: str) -> str:
    ...


@overload
def prompt_ask(prompt: str, default: str | None) -> str | None:
    ...


def prompt_ask(prompt: str, default: str | None = None) -> str | None:
    # this method exists because Prompt.ask text does not parse [markup] tags
    return Prompt.ask(
        _render_as_style(prompt, "promptq"),
        console=_STDERR_CONSOLE,
        # this is how rich does it (strangely)
        default=default if default is not None else ...,  # type: ignore
    )


def confirm_ask(
    prompt: str,
    default: bool | None = None,
) -> bool:
    # ditto
    return Confirm.ask(
        _render_as_style(prompt, "promptq"),
        console=_STDERR_CONSOLE,
        default=default if default is not None else ...,  # type: ignore
    )


def print_json(json_text: str) -> None:
    _STDERR_CONSOLE.print_json(json_text)


if __name__ == "__main__":
    from pathlib import PureWindowsPath

    error("error")
    warn("warn")
    info("info")
    debug("debug")

    print("[path]C:\\Users\\tim\\code\\")
    print(f'[path]{PureWindowsPath("C:/Users/tim/Saved Games/").as_uri()}')
    print("[path]C:/Users/tim/Saved Games/")

    print("[hint]Next time, do this...")
    print("[url]https://rich.readthedocs.io/en/")
    print("this is [package]ThePackage")
    print("this is [addon]MyAddon")

    answer = prompt_ask("How are you?")
    print(f"[prompta]Glad you're doing {answer}")
    confirm_ask("Authorize launch?")

    print("Press [key]Ctrl-C[/key] to exit")
    print("See `[command]wap help[/command]` for more")



import getpass
import re
import textwrap
from pathlib import Path
from typing import Literal

from rich.panel import Panel

from wap.config import AddonConfig, Config, CurseforgeConfig, PublishConfig, TocConfig
from wap.console import confirm_ask, print, print_json, prompt_ask
from wap.exception import AbortError
from wap.wow import FLAVORS, Flavor

_CURSEFORGE_CREATE_PROJECT_LINK = "https://www.curseforge.com/project/1/1/create"


def get_package_name_new_project() -> str:
    while True:
        package_name = prompt_ask("\nPackage name")
        if package_name == "":
            print("[prompta]No package name given. You must provide a package name.")
        elif (Path() / package_name).exists():
            print(
                "[prompta]Package name conflicts with existing file name. Choose "
                "another name."
            )
        else:
            return package_name


def get_package_name_new_config() -> str:
    while True:
        package_name = prompt_ask("\nPackage name")
        if package_name == "":
            print("[prompta]No package name given. You must provide a package name.")
        else:
            return package_name


def get_author() -> str:
    return prompt_ask("\nAuthor", default=getpass.getuser())


def get_desc() -> str:
    return prompt_ask("\nDescription")


def get_version() -> str:
    return prompt_ask("\nVersion", default="0.0.1")


def get_supported_flavors() -> list[Flavor]:
    supported_flavors: list[Flavor] = []
    while True:
        for flavor in FLAVORS:
            if confirm_ask(f"\nSupport {flavor.canon_name}"):
                supported_flavors.append(flavor)
        if supported_flavors:
            return supported_flavors
        print("[prompta]No flavors chosen. You must support at least one flavor.")


def get_project_id() -> str:
    print(
        "\n[hint]If you do not yet have a project, create one at "
        f"[link]{_CURSEFORGE_CREATE_PROJECT_LINK}"
    )
    while True:
        project_id = prompt_ask("Curseforge project ID")
        if not re.fullmatch(r"\d+", project_id):
            print("[prompta]Invalid project ID. It must be a number.")
            continue
        return project_id


def get_project_slug() -> str | None:
    # https://www.curseforge.com/wow/addons/details
    print(
        "\n[hint]Your project slug is in your project's URL. For example, if the URL "
        'is "https://www.curseforge.com/wow/addons/my-addon", then the slug is '
        '"my-addon".'
    )
    return prompt_ask("Curseforge project slug (Press enter to skip)", default=None)


def confirm_publish() -> bool:
    return confirm_ask("\nWill you want to publish to Curseforge?")


def get_publish_config() -> PublishConfig | None:
    if not confirm_publish():
        return None

    project_id = get_project_id()
    project_slug = get_project_slug()

    return PublishConfig(
        curseforge=CurseforgeConfig(
            project_id=project_id,
            slug=project_slug,
            changelog_text="",
        )
    )


def confirm_all_ok() -> bool:
    return confirm_ask("\nIs this OK")


def prompt_for_config(mode: Literal["new-project", "new-config"]) -> Config:
    intro = """\
    These questions will walk you through creating a new wap configuration file.

    It will cover only the most basic settings, and you should edit it later.

    See `[command]wap help[/command]` for more detailed documentation.

    Press [key]Ctrl-C[/key] at any time to quit.\
    """

    print(
        Panel.fit(textwrap.dedent(intro), title=f"wap {mode}", padding=(1, 2), width=80)
    )

    if mode == "new-config":
        package_name = get_package_name_new_config()
    elif mode == "new-project":
        package_name = get_package_name_new_project()

    author = get_author()
    desc = get_desc()
    version = get_version()
    supported_flavors = get_supported_flavors()
    publish_config = get_publish_config()

    config = Config(
        name=package_name,
        version=version,
        author=author,
        wow_versions={
            flavor.name: flavor.latest_version.dotted for flavor in supported_flavors
        },
        publish=publish_config,
        package=[
            AddonConfig(
                path=f"./{package_name}",
                toc=TocConfig(
                    tags={
                        "Title": package_name,
                        "Notes": desc,
                    },
                    files=["Main.lua"],
                ),
            )
        ],
    )

    # i like this for learning reasons. it maps the previous questions to a real config
    print_json(config.to_json())
    if confirm_all_ok():
        return config

    raise AbortError("Aborted!")



from __future__ import annotations

from collections.abc import Mapping, Sequence
from pathlib import Path, PureWindowsPath
from typing import ClassVar

import arrow
from attrs import frozen

from wap import __version__
from wap.config import Config, TocConfig
from wap.exception import PathMissingError, TagError
from wap.wow import Version

# Just some notes about TOC files I found
# - Folder name must match toc prefix. E.g. Foo/Foo.toc, but not Foo/Bar.toc
# - Title not required. If not, addon list shows folder name
# - Localizable tags logic:
#   - For a given player with a given locale named `<locale>`, that player can match
#     to both `Title-<locale>` and `Title` (same for notes). Which one is chosen is
#     based on ordering: last always wins.
#   - Official tag names are case insensitive EXCEPT for locales
#   - Presumably, only Title and Notes may be localized
# - If no Notes are given, title is used
# - Files are optional (but we will require them in our config file)
# - Tags are optional (ditto)
# - Tags do not require values (default to empty string)

# this list comes from some blizzard wow addon code.
# https://www.townlong-yak.com/framexml/beta/VideoOptionsPanels.lua#1270
_LANGUAGE_REGIONS = {
    "deDE",
    "enCN",
    "enGB",
    "enTW",
    "enUS",
    "esES",
    "esMX",
    "frFR",
    "itIT",
    "koKR",
    "ptBR",
    "ptPT",
    "ruRU",
    "zhCN",
    "zhTW",
}

# tag names from https://wowwiki-archive.fandom.com/wiki/TOC_format
_BARE_LOCALIZABLE_TAGS = {"Title", "Notes"}
_LOCALIZED_TAGS = {
    f"{tag}-{lr}" for tag in _BARE_LOCALIZABLE_TAGS for lr in _LANGUAGE_REGIONS
}
_METADATA_TAG_PREFIX = "X-"


@frozen
class Toc:
    tags: Sequence[tuple[str, str]]
    files: Sequence[str]
    flavor_suffix: str | None

    _TAG_LINE_PREFIX: ClassVar[str] = "##"
    _COMMENT_LINE_PREFIX: ClassVar[str] = "#"

    @classmethod
    def parse(cls, contents: str, flavor_suffix: str) -> Toc:
        tags = []
        files = []

        for line in contents.splitlines():
            if line and not line.startswith(cls._COMMENT_LINE_PREFIX):
                files.append(line)

            elif line.startswith(cls._TAG_LINE_PREFIX):
                name, value = [
                    token.strip()
                    for token in line.removeprefix("##").split(sep=":", maxsplit=1)
                ]
                tags.append((name, value))

        return cls(tags=tags, files=files, flavor_suffix=flavor_suffix)

    @classmethod
    def from_toc_config(
        cls,
        toc_config: TocConfig,
        source_path: Path,
        config: Config,
        wow_version: Version,
        flavor_suffix: str | None = None,
    ) -> Toc:
        default_tags = {}
        if config.author is not None:
            default_tags["Author"] = config.author
        if config.description is not None:
            default_tags["Description"] = config.description
        default_tags["Title"] = source_path.name
        default_tags["Version"] = config.version
        default_tags["Interface"] = wow_version.interface_version
        default_tags[
            f"{_METADATA_TAG_PREFIX}BuildDateTime"
        ] = arrow.utcnow().isoformat()
        default_tags[f"{_METADATA_TAG_PREFIX}BuildTool"] = f"wap {__version__}"

        # all default tags are overrideable
        merged_tags = default_tags | toc_config.serialized_tags

        # sort the tags, for Title and Notes
        # TOC parsing logic is that, for a given user locale "xxXX" and tag name "Tag"
        # in {"Title", "Notes"}, the realized value will the LAST declaration of "Tag"
        # or "Tag-xxXX" in the TOC file. If neither are present, the realized value will
        # be the game client's default value. For "Title", the default is the name of
        # the directory, and for "Notes", the default is an empty string.
        # Since locale tags are probably more applicable, we want them last.
        tag_pairs = list(merged_tags.items())
        # this sort should only reorder localized tags. everything else should be in
        # config order (insertion order) because this python sort is stable.
        tag_pairs.sort(key=lambda kv: 1 if kv[0] in _LOCALIZED_TAGS else 0)

        return cls(tags=tag_pairs, files=toc_config.files, flavor_suffix=flavor_suffix)

    @property
    def tag_map(self) -> Mapping[str, str]:
        return dict(self.tags)

    def validate(self, source_dir: Path) -> None:
        """
        Check various things about a toc, including testing that the paths in the files
        list actually exist in source_dir
        """
        illegal_tag_chars = ["\n", " ", ":"]
        for tag in self.tag_map:
            for ill_char in illegal_tag_chars:
                if ill_char in tag:
                    raise TagError(
                        f'Tag {tag} contains illegal character "{ill_char!r}". Please '
                        "remove it and try again"
                    )
        for file_path in self.files:
            joined_path = source_dir / file_path
            if not joined_path.is_file():
                raise PathMissingError(
                    f"TOC file path {joined_path} does not exist. Please fix the path "
                    "in your configuration file or remove it."
                )

    @classmethod
    def _create_tag_line(cls, key: str, value: str) -> str:
        return f"{cls._TAG_LINE_PREFIX} {key}: {value}"

    @staticmethod
    def _create_file_line(path: str) -> str:
        # TOC file use windows path separators
        windows_path = PureWindowsPath(path)
        return f"{windows_path}"

    def filename(self, addon_name: str) -> str:
        return f"{addon_name}{self.flavor_suffix or ''}.toc"

    def generate(self) -> str:
        return "\n".join(
            (
                *(self._create_tag_line(key, value) for key, value in self.tags),
                "",
                *(self._create_file_line(files) for files in self.files),
            )
        )



import textwrap
from pathlib import Path

import click

from wap.console import print
from wap.prompt import prompt_for_config


@click.command()
def new_project() -> None:
    """
    Create a new project.
    """
    config = prompt_for_config(mode="new-project")

    project_path = Path.cwd() / config.name
    project_path.mkdir()

    # create the following structure
    # - README.md
    # - Addon/
    #   - Main.lua
    # - wap.json

    config_path = project_path / "wap.json"
    config.write_to_path(config_path, indent=True)

    readme_md_text = textwrap.dedent(
        f"""\
        # {config.name}

        This addon helps you...
        """
    )
    readme_path = project_path / "README.md"
    readme_path.write_text(readme_md_text)

    main_lua_text = textwrap.dedent(
        f"""\
        -- Your code can go here.
        -- Here's something to get you started, but you can erase it if you wish.

        local title = GetAddOnMetadata("{config.name}", "Title")
        local version = GetAddOnMetadata("{config.name}", "Version")
        print(title .. " version " .. version .. " has loaded")
        """
    )
    main_lua_path = project_path / config.name / "Main.lua"
    main_lua_path.parent.mkdir()
    main_lua_path.write_text(main_lua_text)

    print(
        f"Project for package [package]{project_path.name}[/package] created at "
        f"[path]{project_path}[/path]"
    )



class WapError(Exception):
    """General WAP exception"""

    def __init__(self, message: str):
        super().__init__(message)
        self.message = message


class AbortError(WapError):
    """Indicates the user aborted the operation."""


class ConfigError(WapError):
    """Indicates a general problem with the config."""


class ConfigPathError(WapError):
    """Indicates a problem accessing a configuration value"""


class ConfigValueError(WapError):
    """Indicates a problem encoding a configuration value to JSON"""


class ConfigSchemaError(WapError):
    """Indicates a config does not follow the schema."""


class CurseForgeAPIError(WapError):
    """Indicates a problem communicating with CurseForge"""


class EncodingError(WapError):
    """Indicates an issue encoding or decoding data"""


class PathExistsError(WapError):
    """Indicates a path exists that should not"""


class PathMissingError(WapError):
    """Indicates a needed path does not exist."""


class PathTypeError(WapError):
    """Indicates a path is not of the type expected (directory, file, etc)."""


class PlatformError(WapError):
    """Indicates that the current platform does not have a required feature."""


class TagError(WapError):
    """Indicates a malformed tag inside a TOC."""


class VersionError(WapError):
    """Indicates that a version is invalid when parsed."""



import shutil
from pathlib import Path

import click

from wap.commands.util import (
    DEFAULT_OUTPUT_PATH,
    config_path_option,
    output_path_option,
)
from wap.config import Config
from wap.console import print, warn
from wap.core import get_build_path
from wap.curseforge import RELEASE_TYPES, Changelog, CurseForgeAPI
from wap.exception import ConfigError, CurseForgeAPIError, PathMissingError

DEFAULT_RELEASE_TYPE = "alpha"
WAP_CURSEFORGE_TOKEN_ENVVAR_NAME = "WAP_CURSEFORGE_TOKEN"


@click.command()
@config_path_option()
@output_path_option()
@click.option(
    "-r",
    "--release-type",
    type=click.Choice(list(RELEASE_TYPES)),
    show_default=True,
    help="The type of release to make.",
)
@click.option(
    "--curseforge-token",
    metavar="TOKEN",
    envvar=WAP_CURSEFORGE_TOKEN_ENVVAR_NAME,
    required=True,
    help=(
        "The value of your CurseForge API token. May also be specified in the "
        f"environment variable {WAP_CURSEFORGE_TOKEN_ENVVAR_NAME}."
    ),
)
def publish(
    config_path: Path,
    output_path: Path | None,
    release_type: str | None,
    curseforge_token: str,
) -> None:
    """
    Upload packages to Curseforge.
    """
    config = Config.from_path(config_path)
    if output_path is None:
        output_path = config_path.parent / DEFAULT_OUTPUT_PATH

    if config.publish is None or config.publish.curseforge is None:
        raise ConfigError(
            'A "publish.curseforge" config section should be present to publish. '
            "Please add one and try again."
        )

    cf_config = config.publish.curseforge

    if cf_config.changelog_file is not None:
        changelog_path = config_path.parent / cf_config.changelog_file
        changelog = Changelog.from_path(
            path=changelog_path, type_=cf_config.changelog_type
        )
    elif cf_config.changelog_text is not None:
        changelog = Changelog.from_text(
            text=cf_config.changelog_text, type_=cf_config.changelog_type
        )
    else:
        changelog = Changelog.from_text(text="")
        warn("No changelog text or file provided, so using empty string")

    build_path = get_build_path(output_path, config)
    if not build_path.is_dir():
        raise PathMissingError(
            f'Build path {build_path} should be a directory. Have you run "wap build" '
            "yet?"
        )

    print(f"Zipping [path]{build_path}[path]")
    zip_path = Path(
        shutil.make_archive(
            base_name=str(build_path), format="zip", root_dir=build_path
        )
    )

    cf_api = CurseForgeAPI(api_token=curseforge_token)

    print("Getting CurseForge WoW version ids...")
    version_map = cf_api.get_version_map()

    version_ids = []
    for flavor_name, version in config.wow_versions.items():
        try:
            version_ids.append(version_map[version])
        except KeyError as key_error:
            raise CurseForgeAPIError(
                f"Curseforge does not know about version {version} for flavor "
                f"{flavor_name}. Does it actually exist?"
            ) from key_error

    if release_type is None:
        if cf_config.release_type:
            release_type = cf_config.release_type
        else:
            release_type = DEFAULT_RELEASE_TYPE
            warn(
                "No release type specified on command line or in config. Defaulting to "
                f"{DEFAULT_RELEASE_TYPE}"
            )

    print("Uploading to CurseForge...")
    with zip_path.open("rb") as zip_file:
        file_id = cf_api.upload(
            project_id=cf_config.project_id,
            file=zip_file,
            display_name=build_path.name,  # Addon-1.2.3
            file_name=zip_path.name,  # Addon-1.2.3.zip
            changelog=changelog,
            game_version_ids=version_ids,
            release_type=release_type,
        )
    if cf_config.slug is not None:
        url = cf_api.uploaded_file_url(file_id=file_id, slug=cf_config.slug)
        print(f"Upload available at [url]{url}[url]")
    else:
        print(f"Uploaded file {file_id}")
        print(
            '[hint]Hint: Provide a "slug" in your curseforge config to get an entire '
            "link in output next time."
        )



from __future__ import annotations

from collections.abc import Iterable, Iterator, Sequence
from pathlib import Path
from typing import Literal, cast, get_args

import click
from attrs import frozen
from watchfiles import watch

from wap.commands.util import (
    DEFAULT_OUTPUT_PATH,
    clean_option,
    config_path_option,
    output_path_option,
    wow_addons_dir_options,
)
from wap.config import AddonConfig, Config
from wap.console import print, warn
from wap.core import get_build_path
from wap.exception import ConfigError, PathExistsError, PathTypeError
from wap.fileops import clean_dir, copy_path, delete_path, symlink
from wap.toc import Toc
from wap.wow import FLAVOR_MAP, FLAVOR_NAMES, FlavorName, Version


@frozen(kw_only=True)
class AddonBuildResult:
    path: Path

    def link(self, wow_addons_path: Path, force: bool) -> Path | None:
        """
        Links this built addon to a WoW addons directory. Returns the link path.
        """
        link_path = wow_addons_path / self.path.name

        if (
            force
            and link_path.exists()
            and (link_path.resolve() != self.path.resolve())
        ):
            delete_path(link_path)

        symlink(new_path=link_path, target_path=self.path)

        return link_path


@frozen(kw_only=True)
class Addon:
    source_path: Path
    include_paths: Sequence[Path]
    include_path_root: Path
    tocs: Sequence[Toc]

    @classmethod
    def create(
        cls,
        addon_config: AddonConfig,
        config: Config,
        config_path: Path,
    ) -> Addon:
        config_dir = config_path.parent

        source_path = (config_dir / addon_config.path).resolve()

        include_paths = (
            resolve_globs(root_path=config_dir, glob_patterns=addon_config.include)
            if addon_config.include
            else []
        )

        tocs = []
        if addon_config.toc is not None:
            wow_versions = []
            for flavor_name, flavor_version in config.wow_versions.items():
                wow_version = Version.from_dotted(flavor_version)
                toc = Toc.from_toc_config(
                    toc_config=addon_config.toc,
                    wow_version=wow_version,
                    flavor_suffix=FLAVOR_MAP[flavor_name].toc_suffix,
                    config=config,
                    source_path=source_path,
                )
                tocs.append(toc)
                wow_versions.append(wow_version)

            # create just a standard toc with no suffix because curseforge purportedly
            # requires it. just use the max wow_version from those provided.
            max_wow_version = max(wow_versions)
            tocs.append(
                Toc.from_toc_config(
                    toc_config=addon_config.toc,
                    wow_version=max_wow_version,
                    config=config,
                    source_path=source_path,
                )
            )

        return cls(
            source_path=source_path,
            tocs=tocs,
            include_paths=include_paths,
            include_path_root=config_path.parent,
        )

    def build(self, package_path: Path, clean: bool) -> AddonBuildResult:
        build_path = package_path / self.source_path.name

        try:
            build_path.mkdir(parents=True, exist_ok=True)
        except FileExistsError as file_exists_error:
            raise PathExistsError(
                f"Output directory {build_path} should not be a file"
            ) from file_exists_error

        if clean:
            clean_dir(build_path)

        if not self.source_path.is_dir():
            raise PathTypeError(f"Addon path {self.source_path} should be a directory.")

        # copy source to dest
        copy_path(self.source_path, build_path)

        # keep track of what we've built
        addon_paths = {
            path.relative_to(self.source_path) for path in self.source_path.rglob("*")
        }

        # copy includes
        for include_path in self.include_paths:
            include_path_target = build_path / include_path.name
            rel_path = Path(include_path.name)
            if rel_path in addon_paths:
                warn(f"Include path {rel_path} already exists in output directory")
            else:
                addon_paths.add(rel_path)
            copy_path(include_path, include_path_target)

        # write tocs
        toc_paths = []
        for toc in self.tocs:
            toc.validate(build_path)
            toc_path_target = build_path / toc.filename(build_path.name)
            rel_path = Path(toc_path_target.name)
            if rel_path in addon_paths:
                warn(
                    f"Generated TOC path {rel_path} already exists in output "
                    "directory"
                )
            else:
                addon_paths.add(rel_path)
            try:
                toc_path_target.write_text(toc.generate())
            except (PermissionError, IsADirectoryError) as error:
                # on windows, raises PermissionError, linux raises IsADirectoryError
                raise PathExistsError(
                    f"Generated TOC file {toc_path_target} should not exist in your "
                    "source files."
                ) from error
            toc_paths.append(toc_path_target)

        return AddonBuildResult(path=build_path)

    @property
    def watch_paths(self) -> Sequence[Path]:
        return [self.source_path, *self.include_paths]


@frozen(kw_only=True)
class Package:
    addons: Sequence[Addon]
    build_path: Path

    @classmethod
    def create(cls, config: Config, config_path: Path, output_path: Path) -> Package:
        package = cls(
            addons=[
                Addon.create(
                    addon_config=addon_config,
                    config=config,
                    config_path=config_path,
                )
                for addon_config in config.package
            ],
            build_path=get_build_path(output_path=output_path, config=config),
        )

        # dupe check
        seen_addon_paths = set()
        for addon in package.addons:
            if addon.source_path not in seen_addon_paths:
                seen_addon_paths.add(addon.source_path)
            else:
                raise ConfigError(
                    "Addon paths should be unique. Found duplicate for "
                    f"{addon.source_path}."
                )

        return package

    def build(self, clean: bool) -> Sequence[AddonBuildResult]:
        return [
            addon.build(package_path=self.build_path, clean=clean)
            for addon in self.addons
        ]

    @property
    def watch_paths(self) -> Sequence[Path]:
        return [watch_path for addon in self.addons for watch_path in addon.watch_paths]


AutoChoiceName = Literal["auto"]
AUTO_CHOICE: AutoChoiceName = get_args(AutoChoiceName)[0]


def get_addon_link_targets(
    flavors_to_link: Sequence[FlavorName | AutoChoiceName],
    config: Config,
    mainline_addons_path: Path,
    wrath_addons_path: Path,
    vanilla_addons_path: Path,
) -> dict[FlavorName, Path]:
    flavor_addons_path_map: dict[FlavorName, Path] = {
        "mainline": mainline_addons_path,
        "wrath": wrath_addons_path,
        "vanilla": vanilla_addons_path,
    }
    uniq_flavors = set(flavors_to_link)

    if AUTO_CHOICE in uniq_flavors:
        if len(uniq_flavors) > 1:
            raise click.BadOptionUsage(
                "link",
                ('If linking "--auto", it should be the only provided link option.'),
            )

        config_flavors = set(config.wow_versions)
        existing_installs = set(
            flavor_name
            for flavor_name in FLAVOR_NAMES
            if flavor_addons_path_map[flavor_name].exists()
        )
        return {
            flavor_name: flavor_addons_path_map[flavor_name]
            for flavor_name in config_flavors & existing_installs
        }

    return {
        link_flavor: flavor_addons_path_map[link_flavor]
        for link_flavor in cast(set[FlavorName], uniq_flavors)
    }


def resolve_globs(root_path: Path, glob_patterns: Sequence[str]) -> Sequence[Path]:
    paths = []
    for pattern in glob_patterns:
        matching_paths = list(root_path.glob(pattern))
        if len(matching_paths) == 0:
            warn(
                f'Include pattern "{pattern}" should match some paths, but none were '
                "found."
            )
        paths.extend(matching_paths)
    return paths


@click.command()
@config_path_option()
@output_path_option()
@clean_option()
@click.option(
    "-l",
    "--link",
    "flavors_to_link",
    type=click.Choice([AUTO_CHOICE, *FLAVOR_NAMES], case_sensitive=False),
    multiple=True,
    is_flag=False,
    flag_value=AUTO_CHOICE,
    help=(
        """
        Create a symlink from packaged addon folder into a flavor's WoW AddOns. This is
        handy for installing your own stuff, so you can work on it and test it quickly.
        If no argument is provided to this option or if "auto" is provided, then links
        will be made into each flavor that exists on this system and is supported by the
        package. This option can be provided mulitple times.
        """
    ),
    show_default=True,
)
@click.option(
    "--link-force",
    is_flag=True,
    help=(
        "If --link and the link path already exists, delete it first so that this link "
        "can be made."
    ),
)
@click.option(
    "-w",
    "--watch",
    "enable_watch",
    is_flag=True,
    help=("Repackage when source files change"),
)
@wow_addons_dir_options()
def build(
    config_path: Path,
    output_path: Path | None,
    clean: bool,
    flavors_to_link: Sequence[FlavorName | AutoChoiceName],
    link_force: bool,
    enable_watch: bool,
    mainline_addons_path: Path,
    wrath_addons_path: Path,
    vanilla_addons_path: Path,
) -> None:
    """
    Build addons into a playable, distributable package.
    """
    config_path = config_path.resolve()

    if output_path is None:
        output_path = config_path.parent / DEFAULT_OUTPUT_PATH

    # used to signal if this is the first time we're building, where we might print more
    # information that subsequent times (in watch mode).
    first_time = True

    def build_once() -> Package:
        config = Config.from_path(config_path)
        package = Package.create(
            config=config,
            config_path=config_path,
            output_path=output_path,  # type: ignore
            # mypy bug https://github.com/python/mypy/issues/2608
        )

        built_addons = package.build(clean=clean)

        addon_link_dirs = get_addon_link_targets(
            flavors_to_link,
            config,
            mainline_addons_path=mainline_addons_path,
            wrath_addons_path=wrath_addons_path,
            vanilla_addons_path=vanilla_addons_path,
        )

        for addon in built_addons:
            build_addon_msg = f"Built addon [addon]{addon.path.name}[/addon]"
            if first_time:
                build_addon_msg += f" at [path]{addon.path}[/path]"
            print(build_addon_msg)

            for flavor_name, addon_dir in addon_link_dirs.items():
                link_path = addon.link(wow_addons_path=addon_dir, force=link_force)
                if first_time:
                    print(
                        f"Linked [path]{link_path}[/path] "
                        f"([flavor]{flavor_name}[/flavor]) to "
                        f"[addon]{addon.path.name}[/addon]"
                    )

        build_package_msg = (
            f"Built package [package]{package.build_path.name}[/package]"
        )
        if first_time:
            build_package_msg += f" at [path]{package.build_path}[/path]"
        print(build_package_msg)

        return package

    package = build_once()
    first_time = False

    if enable_watch:
        print("Running in watch mode. Press [key]Ctrl-C[/key] at any time to quit.")
        project_file_paths = {*package.watch_paths, config_path}
        for paths_changed in watch_paths(config_path.parent):
            if any(
                changed_path.is_relative_to(watch_path)
                for watch_path in project_file_paths
                for changed_path in paths_changed
            ):
                print("Project file changed, rebuilding...\n")
                package = build_once()
                project_file_paths = {*package.watch_paths, config_path}


def watch_paths(*paths: Path) -> Iterator[Iterable[Path]]:
    for changes in watch(*paths):
        yield {Path(path) for _, path in changes}



from pathlib import Path

import click

from wap.commands.util import config_path_option
from wap.config import Config
from wap.console import error, print
from wap.exception import WapError


@click.command()
@config_path_option()
def validate(
    config_path: Path,
) -> int:
    """
    Validate a configuration file.
    """
    try:
        Config.from_path(config_path)
    except WapError as wap_exc:
        error(wap_exc.message)
        print("invalid", stderr=False)
        return 1

    print("valid", stderr=False)
    return 0



import webbrowser
from textwrap import dedent

import click

from wap import __name__ as package_name
from wap import __version__
from wap.commands.build import build
from wap.commands.new_config import new_config
from wap.commands.new_project import new_project
from wap.commands.publish import publish
from wap.commands.validate import validate
from wap.console import print


def open_or_print_help_url(command_name: str | None = None) -> None:
    if command_name is None:
        full_url = BASE_HELP_URL
    else:
        full_url = f"{BASE_HELP_URL}/commands/{command_name}"

    try:
        webbrowser.open(full_url)
    except webbrowser.Error:
        print(f"See documentation at [url]{full_url}[/url]")


# this subcommand is defined here because it uses the list of all subcommands. if it
# were in its own file (like the other subcommands), there'd be a circular dependency.
@click.command("help")
@click.argument("subcommand", required=False)
def help_command(subcommand: str | None) -> None:
    """Open the help webpage for the given SUBCOMMAND in your browser."""
    if subcommand is None:
        open_or_print_help_url()
    elif subcommand not in SUBCOMMAND_NAMES:
        raise click.BadArgumentUsage(
            f"Unknown subcommand {subcommand}. See available subcommands with "
            "`wap help`."
        )
    else:
        open_or_print_help_url(subcommand)


SUBCOMMANDS = [
    build,
    help_command,
    new_config,
    new_project,
    publish,
    validate,
]
SUBCOMMAND_NAMES = {command.name for command in SUBCOMMANDS}
BASE_HELP_URL = "http://t-mart.github.io/wap"


@click.group()
@click.version_option(
    package_name=package_name,
    message=f"wap version {__version__}",
)
def base() -> int | None:
    """World of Warcraft addon packager"""


for subcommand in SUBCOMMANDS:
    base.add_command(subcommand)
    if subcommand.help:
        subcommand.help = (
            dedent(subcommand.help)
            + f"\n\nRun `wap help {subcommand.name}` for more information."
        )



from __future__ import annotations

import importlib.resources
import json
import re
import sys
from pathlib import Path
from typing import ClassVar, Literal, Mapping, get_args

from attrs import field, frozen

import wap
from wap.exception import VersionError


@frozen(order=True)
class Version:

    _INTERFACE_PATTERN: ClassVar[
        str
    ] = r"""(?x)
        ^
        (?P<major>\d+)
        (?P<minor>\d{2})
        (?P<patch>\d{2})
        $
    """
    _DOT_PATTERN: ClassVar[
        str
    ] = r"""(?x)
        ^
        (?P<major>\d+)
        \.
        (?P<minor>\d+)
        \.
        (?P<patch>\d+)
        $
    """

    major: int
    minor: int
    patch: int

    @classmethod
    def from_dotted(cls, dotted: str) -> Version:
        if match := re.fullmatch(cls._DOT_PATTERN, dotted):
            return Version(
                major=int(match.group("major")),
                minor=int(match.group("minor")),
                patch=int(match.group("patch")),
            )
        raise VersionError(f"Dotted version {dotted} does not appear to be valid.")

    @classmethod
    def from_interface(cls, interface: str) -> Version:
        if match := re.fullmatch(cls._INTERFACE_PATTERN, interface):
            return Version(
                major=int(match.group("major")),
                minor=int(match.group("minor")),
                patch=int(match.group("patch")),
            )
        raise VersionError(
            f"Interface version {interface} does not appear to be valid."
        )

    @property
    def dotted(self) -> str:
        return f"{self.major}.{self.minor}.{self.patch}"

    @property
    def interface_version(self) -> str:
        return f"{self.major}{self.minor:0>2}{self.patch:0>2}"


FlavorName = Literal[
    "mainline",
    "wrath",
    "vanilla",
]
FLAVOR_NAMES: tuple[FlavorName, ...] = get_args(FlavorName)


@frozen(kw_only=True)
class Flavor:
    name: FlavorName
    canon_name: str
    toc_suffix: str
    latest_version: Version = field(eq=False)
    installation_dir_name: str


def _get_latest_versions() -> Mapping[FlavorName, str]:
    path = importlib.resources.files(wap).joinpath("versions/latest.json")
    data = path.read_text(encoding="utf-8")
    obj: Mapping[FlavorName, str] = json.loads(data)
    return obj


_latest_versions = _get_latest_versions()

MAINLINE_FLAVOR = Flavor(
    name="mainline",
    canon_name="Mainline",
    toc_suffix="_Mainline",
    latest_version=Version.from_dotted(_latest_versions["mainline"]),
    installation_dir_name="_retail_",
)
WRATH_FLAVOR = Flavor(
    name="wrath",
    canon_name="Wrath of the Lich King Classic",
    toc_suffix="_Wrath",
    latest_version=Version.from_dotted(_latest_versions["wrath"]),
    installation_dir_name="_classic_",
)
VANILLA_FLAVOR = Flavor(
    name="vanilla",
    canon_name="WoW Classic",
    toc_suffix="_Vanilla",
    latest_version=Version.from_dotted(_latest_versions["vanilla"]),
    installation_dir_name="_classic_era_",
)

FLAVORS = [
    MAINLINE_FLAVOR,
    WRATH_FLAVOR,
    VANILLA_FLAVOR,
]

FLAVOR_MAP = {flavor.name: flavor for flavor in FLAVORS}

_DEFAULT_INSTALL_PATH_FOR_PLATFORM = {
    "win32": R"C:\Program Files (x86)\World of Warcraft",
    "darwin": "/Applications/World of Warcraft",
}


def get_default_addons_path(flavor: Flavor, platform: str | None = None) -> Path | None:
    if not platform:
        platform = sys.platform

    if platform not in _DEFAULT_INSTALL_PATH_FOR_PLATFORM:
        return None

    return (
        Path(_DEFAULT_INSTALL_PATH_FOR_PLATFORM[platform])
        / flavor.installation_dir_name
        / "Interface"
        / "AddOns"
    )



import shutil
from pathlib import Path

from wap.exception import (
    PathExistsError,
    PathMissingError,
    PathTypeError,
    PlatformError,
)


def delete_path(path: Path) -> None:
    """
    Deletes either a file or directory. The path existance must be checked before
    calling this function.
    """
    if path.is_file():
        path.unlink()
    elif path.is_dir():
        shutil.rmtree(path)
    else:  # pragma: no cover
        raise PathTypeError(
            f"Path {path} should be a file or directory in order to delete it."
        )


def clean_dir(path: Path) -> None:
    """
    Deletes all files and subdirectories in a given directory, but does not delete the
    directory itself.
    """
    for subpath in path.iterdir():
        delete_path(subpath)


def copy_path(src: Path, dst: Path) -> None:
    """
    If src is a file: If dst does not exist, copy src to dst. If dst is a file,
    overwrite its contents with those of src. If dst is a directory, raise a
    PathExistsException.

    If src is a directory: If dst does not exist, create a new directory dst and copy
    src's tree to it. If dst is a directory, copy src's tree to it (without cleaning).
    If dst is a file, raise a PathExistsException.

    | src type | dest type |    outcome    |
    |:--------:|:---------:|:-------------:|
    |     F    |     F     |   overwrite   |
    |     F    |     D     |  exists error |
    |     F    | not exist |    new file   |
    |     D    |     F     |   perm error  |
    |     D    |     D     | copy contents |
    |     D    | not exist |    new dir    |

    And finally, if src and dst are not directories or files, raise a PathTypeException.
    """
    if src.is_file():
        if dst.is_dir():
            raise PathExistsError(
                f"Cannot copy file {src} to {dst} because it is a directory. Please "
                "remove that directory or choose a different target."
            )
        shutil.copy(src, dst)
    elif src.is_dir():
        if dst.is_file():
            raise PathExistsError(
                f"Cannot copy directory {src} to {dst} because it is a file. Please "
                "remove that file or choose a different target."
            )
        shutil.copytree(src, dst, dirs_exist_ok=True)
    else:  # pragma: no cover
        raise PathTypeError(
            f"Cannot copy path {src} because it is not a file or directory."
        )


def symlink(
    *, new_path: Path, target_path: Path, target_is_directory: bool | None = None
) -> None:
    """
    Make new_path a symbolic (soft) link that targets target_path.

    On Windows, *target_is_directory* must be set to True if target is a directory. If
    the argument is None (the default), it will take the value of
    `target_path.is_dir()`.

    Additionally on Windows, if a certain type of OSError is raised that indicates
    insufficient privileges, this function instead raises a new OSError (`from` the
    original) with a more helpful message about disabling the privilege requirement'
    for newer versions of Windows 10.
    """
    if target_is_directory is None:
        target_is_directory = target_path.is_dir()

    # FileExistsError and FileNotFoundError are subclasses of OSError, all three of
    # which we catch, so the ordering is important.
    # additionally, target_is_directory is only important in specific Windows cases, but
    # is ignored otherwise, so we provide it as an argument anyway.
    try:
        new_path.symlink_to(
            target_path.resolve(), target_is_directory=target_is_directory
        )
    except FileExistsError as file_exists_error:
        if target_path.resolve() == new_path.resolve():
            return
        raise PathExistsError(
            f"Intended symlink {new_path} already exists and points to a different "
            'target. Delete the symlink and try again, or provide the "--link-force" '
            "option to delete it automatically."
        ) from file_exists_error
    except FileNotFoundError as file_not_found_error:
        raise PathMissingError(
            f"Unable to create link {new_path}. Ensure its parent directory exists."
        ) from file_not_found_error
    except OSError as os_error:
        if getattr(os_error, "winerror", None) == 1314:
            raise PlatformError(
                f"Could not create symbolic link from {new_path} to {target_path} "
                "because Windows Developer Mode is not enabled or because "
                "this program is not running as an administrator."
            ) from os_error
        raise os_error



"""
Main entrypoint to the module
"""
import sys

import click

from wap.commands.base import base
from wap.console import error
from wap.exception import WapError


def main() -> None:
    try:
        # standalone causes ClickExceptions to bubble here, where we will catch them
        # otherwise, ugly stack trace
        exit_code: int | None = base.main(standalone_mode=False, color=True)
        if exit_code is None:
            exit_code = 0
    except WapError as wap_exc:
        error(wap_exc.message)
        exit_code = 1
    except click.ClickException as click_exc:
        error(str(click_exc))
        exit_code = 1
    except click.Abort:
        error("Aborted!")
        exit_code = 130  # http://tldp.org/LDP/abs/html/exitcodes.html

    sys.exit(exit_code)


if __name__ == "__main__":
    main()



from pathlib import Path

import click

from wap.commands.util import DEFAULT_CONFIG_PATH
from wap.console import print
from wap.prompt import prompt_for_config


@click.command()
@click.option(
    "-c",
    "--config-path",
    type=click.Path(file_okay=False, dir_okay=False, path_type=Path),
    default=str(DEFAULT_CONFIG_PATH),
    show_default=True,
    help=("The path to the new configuration file."),
)
def new_config(
    config_path: Path,
) -> None:
    """
    Create a new configuration file in the current directory.
    """
    config = prompt_for_config(mode="new-config")

    config.write_to_path(config_path, indent=True)

    print(f"Created config file at [path]{config_path.resolve()}[path]")



from __future__ import annotations

import importlib.resources
import json
from collections.abc import Mapping, Sequence
from pathlib import Path
from typing import Any, Callable

import jsonschema
from attrs import field, frozen

import wap
from wap.curseforge import ChangelogType, ReleaseType
from wap.wow import FlavorName

from .exception import ConfigSchemaError, EncodingError

SCHEMA_URL = (
    "https://raw.githubusercontent.com/t-mart/wap/master/src/wap/schema/wap.schema.json"
)


def get_schema() -> Any:
    path = importlib.resources.files(wap).joinpath("schema/wap.schema.json")
    data = path.read_text(encoding="utf-8")
    return json.loads(data)


@frozen(kw_only=True)
class Config:
    name: str
    version: str
    author: str | None = field(default=None)
    description: str | None = field(default=None)
    wow_versions: Mapping[FlavorName, str]
    publish: PublishConfig | None = field(default=None)
    package: Sequence[AddonConfig]

    @classmethod
    def from_python_object(cls, obj: Mapping[str, Any]) -> Config:
        try:
            jsonschema.validate(obj, get_schema())
        except jsonschema.ValidationError as validation_error:
            raise ConfigSchemaError(
                f"Invalid configuation: {validation_error.message} at path "
                f"{validation_error.json_path}. Please correct your configuration and "
                "try again."
            ) from validation_error

        publish_obj = obj.get("publish", None)
        if publish_obj is not None:
            publish = PublishConfig.from_python_object(publish_obj)
        else:
            publish = None

        return cls(
            name=obj["name"],
            version=obj["version"],
            author=obj.get("author", None),
            wow_versions=obj["wowVersions"],
            publish=publish,
            package=[AddonConfig.from_python_object(obj) for obj in obj["package"]],
        )

    @classmethod
    def from_path(cls, path: Path) -> Config:
        try:
            return cls.from_python_object(json.loads(path.read_text(encoding="utf-8")))
        except UnicodeDecodeError as unicode_decode_error:
            raise EncodingError(
                f'Config file "{path}" should be utf-8: {unicode_decode_error}'
            ) from unicode_decode_error
        except json.JSONDecodeError as json_decode_error:
            raise EncodingError(
                f'Config file "{path}" should be well-formed JSON: {json_decode_error}.'
            ) from json_decode_error

    def to_python_object(self, with_schema: bool = True) -> Any:
        # i like a certain key order
        obj: dict[str, Any] = {}
        if with_schema:
            obj["$schema"] = SCHEMA_URL
        obj["name"] = self.name
        obj["version"] = self.version
        if self.author is not None:
            obj["author"] = self.author
        if self.description is not None:
            obj["description"] = self.description
        obj["wowVersions"] = self.wow_versions
        if self.publish is not None:
            obj["publish"] = self.publish.to_python_object()
        obj["package"] = [addon.to_python_object() for addon in self.package]
        obj["name"] = self.name

        try:
            jsonschema.validate(obj, get_schema())
        except jsonschema.ValidationError as validation_error:
            raise ConfigSchemaError(
                f"Invalid configuation: {validation_error.message} at path "
                f"{validation_error.json_path}  Please correct your configuration and "
                "try again."
            ) from validation_error

        return obj

    def to_json(self, with_schema: bool = True, indent: bool = False) -> str:
        return json.dumps(
            self.to_python_object(with_schema=with_schema),
            indent=2 if indent else None,
        )

    def write_to_path(
        self, path: Path, with_schema: bool = True, indent: bool = False
    ) -> None:
        path.write_text(self.to_json(with_schema=with_schema, indent=indent))


@frozen(kw_only=True)
class PublishConfig:
    curseforge: CurseforgeConfig | None = field(default=None)

    @classmethod
    def from_python_object(cls, obj: Mapping[str, Any]) -> PublishConfig:
        curseforge_obj = obj.get("curseforge", None)
        if curseforge_obj is not None:
            curseforge = CurseforgeConfig.from_python_object(curseforge_obj)
        else:
            curseforge = None

        return cls(curseforge=curseforge)

    def to_python_object(self) -> dict[str, Any]:
        if self.curseforge is not None:
            return {"curseforge": self.curseforge.to_python_object()}
        return {}


@frozen(kw_only=True)
class CurseforgeConfig:
    project_id: str
    slug: str | None = field(default=None)
    changelog_file: str | None = field(default=None)
    changelog_text: str | None = field(default=None)
    changelog_type: ChangelogType | None = field(default=None)
    release_type: ReleaseType | None = field(default=None)

    @classmethod
    def from_python_object(cls, obj: Mapping[str, Any]) -> CurseforgeConfig:
        return cls(
            project_id=obj["projectId"],
            slug=obj.get("slug", None),
            changelog_file=obj.get("changelogFile", None),
            changelog_text=obj.get("changelogText", None),
            changelog_type=obj.get("changelogType", None),
            release_type=obj.get("releaseType", None),
        )

    def to_python_object(self) -> dict[str, Any]:
        obj: dict[str, Any] = {}
        obj["projectId"] = self.project_id
        if self.slug:
            obj["slug"] = self.slug
        if self.changelog_file:
            obj["changelogFile"] = self.changelog_file
        if self.changelog_text:
            obj["changelogText"] = self.changelog_text
        if self.changelog_type:
            obj["changelogType"] = self.changelog_type
        if self.release_type:
            obj["releaseType"] = self.release_type
        return obj


@frozen(kw_only=True)
class AddonConfig:
    path: str
    toc: TocConfig | None = field(default=None)
    include: Sequence[str] | None = field(default=None)

    @classmethod
    def from_python_object(cls, obj: Mapping[str, Any]) -> AddonConfig:
        toc_obj = obj.get("toc", None)
        if toc_obj is not None:
            toc = TocConfig.from_python_object(toc_obj)
        else:
            toc = None

        return cls(
            path=obj["path"],
            toc=toc,
            include=obj.get("include", None),
        )

    def to_python_object(self) -> dict[str, Any]:
        obj: dict[str, Any] = {}
        obj["path"] = self.path
        if self.toc:
            obj["toc"] = self.toc.to_python_object()
        if self.include:
            obj["include"] = self.include
        return obj


@frozen(kw_only=True)
class TocConfig:
    tags: Mapping[str, bool | str | Sequence[str]]
    files: Sequence[str]

    @classmethod
    def from_python_object(cls, obj: Mapping[str, Any]) -> TocConfig:
        return cls(
            tags=obj.get("tags", {}),
            files=obj.get("files", []),
        )

    @property
    def serialized_tags(self) -> Mapping[str, str]:
        def enabled(value: bool) -> str:
            return "enabled" if value else "disabled"

        def zero_one(value: bool) -> str:
            return "1" if value else "0"

        transformer: dict[str, Callable[[Any], str]] = {
            "LoadOnDemand": zero_one,
            "Dependencies": ", ".join,
            "OptionalDeps": ", ".join,
            "LoadWith": ", ".join,
            "LoadManagers": ", ".join,
            "DefaultState": enabled,
            "SavedVariables": ", ".join,
            "SavedVariablesPerCharacter": ", ".join,
        }

        def pass_through(value: str) -> str:
            return value

        return {
            key: transformer.get(key, pass_through)(value)
            for key, value in self.tags.items()
        }

    def to_python_object(self) -> dict[str, Any]:
        obj: dict[str, Any] = {}
        obj["tags"] = self.tags
        obj["files"] = self.files
        return obj



from collections.abc import Callable
from functools import update_wrapper
from pathlib import Path
from typing import Literal, ParamSpec, TypeVar

import click

from wap.wow import FLAVORS, get_default_addons_path

DEFAULT_OUTPUT_PATH = Path("dist")
DEFAULT_CONFIG_PATH = Path("wap.json")
DISCOVER_SENTINEL = object()

P = ParamSpec("P")
T = TypeVar("T")


class DiscoveredConfigPath(click.ParamType):
    # this gives is git-like discovery of the wap.json file, where this directory and
    # its parents are searched until wap.json is found. if it is not found, fail.

    # CAREFUL when testing this on a real filesystem that may have a wap.json somewhere
    # the in the test's directory heirarchy.
    name = "file"

    def convert(
        self,
        value: str | Literal[False],
        param: click.Parameter | None,
        ctx: click.Context | None,
    ) -> Path:
        if value:
            path = Path(value)
            if not path.is_file():
                self.fail(f"{value} is not a file path", param, ctx)
            return path

        cwd = Path().resolve()

        for parent in [cwd, *cwd.parents]:
            config_path = parent / DEFAULT_CONFIG_PATH
            if config_path.is_file():
                return config_path

        self.fail(
            f'could not find config file "{DEFAULT_CONFIG_PATH}" here or in any parent '
            "directories",
            param,
            ctx,
        )


def config_path_option() -> Callable[[Callable[P, T]], Callable[P, T]]:
    def wrapper(func: Callable[P, T]) -> Callable[P, T]:
        decorated = click.option(
            "-c",
            "--config-path",
            type=DiscoveredConfigPath(),
            default=False,
            show_default=f"{DEFAULT_CONFIG_PATH} in this or a parent directory",
            help=("The path to the configuration file."),
        )(func)

        return update_wrapper(decorated, func)

    return wrapper


def output_path_option() -> Callable[[Callable[P, T]], Callable[P, T]]:
    def wrapper(func: Callable[P, T]) -> Callable[P, T]:
        decorated = click.option(
            "-o",
            "--output-path",
            default=None,
            show_default=f"{DEFAULT_OUTPUT_PATH} directory where config is located",
            type=click.Path(file_okay=False, path_type=Path),
            help=("Root directory path for built packages."),
        )(func)

        return update_wrapper(decorated, func)

    return wrapper


def clean_option() -> Callable[[Callable[P, T]], Callable[P, T]]:
    def wrapper(func: Callable[P, T]) -> Callable[P, T]:
        decorated = click.option(
            "--clean",
            is_flag=True,
            help=(
                "Clean addon directories of files and subdirectories before packaging."
            ),
        )(func)

        return update_wrapper(decorated, func)

    return wrapper


def wow_addons_dir_options() -> Callable[[Callable[P, T]], Callable[P, T]]:
    def wrapper(func: Callable[P, T]) -> Callable[P, T]:
        decorated = func
        for flavor in FLAVORS:
            decorated = click.option(
                f"--{flavor.name}-addons-path",
                type=click.Path(file_okay=False, path_type=Path),
                default=get_default_addons_path(flavor),
                help=(
                    f"Path to the AddOns folder of your {flavor.canon_name} "
                    "installation"
                ),
                show_default=True,
            )(decorated)

        return update_wrapper(decorated, func)

    return wrapper



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['wap', 'wap.commands']

package_data = \
{'': ['*'], 'wap': ['schema/*', 'versions/*']}

install_requires = \
['arrow>=1.2.3,<2.0.0',
 'attrs>=22.1.0,<23.0.0',
 'click>=8.1.3,<9.0.0',
 'httpx>=0.23.0,<0.24.0',
 'jsonschema>=4.15.0,<5.0.0',
 'rich>=12.5.1,<13.0.0',
 'watchfiles>=0.16.1,<0.17.0']

entry_points = \
{'console_scripts': ['wap = wap.__main__:main']}

setup_kwargs = {
    'name': 'wow-addon-packager',
    'version': '0.10.8',
    'description': 'A developer-friendly World of Warcraft addon packager',
    'long_description': '# wap\n\n[![GitHub Actions CI status for default branch](https://github.com/t-mart/wap/actions/workflows/ci.yml/badge.svg)](https://github.com/t-mart/wap/actions/workflows/ci.yml)\n[![Latest release on PyPI](https://img.shields.io/pypi/v/wow-addon-packager)](https://pypi.org/project/wow-addon-packager/)\n\nA developer-friendly World of Warcraft addon packager.\n\n![demonstration of wap usage](https://raw.githubusercontent.com/t-mart/wap/master/docs/assets/demo.gif)\n\n## Features\n\n- [Builds](https://t-mart.github.io/wap/commands/build/) Retail, Wrath, Classic addons (or all\n  three!)\n- [Publishes](https://t-mart.github.io/wap/commands/publish/) your addons to CurseForge\n- [Generates valid TOC files](https://t-mart.github.io/wap/toc-gen/) automagically\n- [Continuously rebuilds](https://t-mart.github.io/wap/commands/build/#-watch) your addon during\n  development\n- Sets up new addon projects quickly, ready to go with one\n  [command](https://t-mart.github.io/wap/commands/new-project/)\n- Consolidates all configuration in\n  [one easy-to-edit file](https://t-mart.github.io/wap/configuration)\n- Supports and is tested on Windows, macOS, and Linux\n- Has awesome [documentation](https://t-mart.github.io/wap/)\n\n## wap in 5 minutes\n\nThese instructions create and upload a working addon without editing a single line of code!\n\n1. Download and install [Python 3.11](https://www.python.org/downloads/).\n\n2. Install `wap`:\n\n    ```console\n    pip install --upgrade --user wow-addon-packager\n    ```\n\n3. Create a new a project:\n\n    ```console\n    wap new-project\n    ```\n\n    And then, answer the prompts. Don\'t worry too much about your answers -- you can always change\n    them later in your configuration file.\n\n4. Change to your new project\'s directory. For example, if you named it `MyAddon` in the last step,\n   you\'d type:\n\n    ```console\n    cd MyAddon\n    ```\n\n5. Build your addon package and link it to your local World of Warcraft installation:\n\n    ```console\n    wap build --link\n    ```\n\n    At this point, **you can play the game with your addon**.\n\n6. Upload your addon to CurseForge with your\n   [API token](https://authors.curseforge.com/account/api-tokens) so that others can use it:\n\n    ```console\n    wap publish --curseforge-token "<api-token>"\n    ```\n\n## Project Information\n\n- License: MIT\n- PyPI: <https://pypi.org/project/wow-addon-packager/>\n- Source Code: <https://github.com/t-mart/wap>\n- Documentation: <https://t-mart.github.io/wap/>\n- Supported Python Versions: 3.11 and later\n- Badge: [![Packaged by wap](https://img.shields.io/badge/packaged%20by-wap-d33682)](https://github.com/t-mart/wap)\n- Contribution Guide: <https://t-mart.github.io/wap/contributing>\n',
    'author': 'Tim Martin',
    'author_email': 'tim@timmart.in',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/t-mart/wap',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.11,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

import json
from collections.abc import Mapping, Sequence
from pathlib import Path
from typing import BinaryIO, ClassVar, Literal, NewType, get_args

import httpx
from attrs import frozen

from wap.console import warn
from wap.exception import CurseForgeAPIError, EncodingError, PathMissingError

GameVersionId = NewType("GameVersionId", int)
ChangelogType = Literal["text", "html", "markdown"]
CHANGELOG_TYPES: tuple[ChangelogType, ...] = get_args(ChangelogType)
ReleaseType = Literal["alpha", "beta", "release"]
RELEASE_TYPES: tuple[ReleaseType, ...] = get_args(ReleaseType)


def _raise_for_status(response: httpx.Response, activity_text: str) -> None:
    if response.status_code != httpx.codes.OK:
        raise CurseForgeAPIError(
            f"HTTP Error from {response.url} during {activity_text}: "
            f"'{response.status_code} {response.reason_phrase}'. Response body: "
            f"{response.text}."
        )


@frozen(kw_only=True)
class CurseForgeAPI:

    api_token: str

    _CLIENT: ClassVar[httpx.Client] = httpx.Client()
    TOKEN_HEADER_NAME: ClassVar[str] = "X-Api-Token"
    UPLOADED_FILE_URL_TEMPLATE: ClassVar[
        str
    ] = "https://www.curseforge.com/wow/addons/{slug}/files/{file_id}"

    VERSION_ENDPOINT_URL: ClassVar[str] = "https://wow.curseforge.com/api/game/versions"
    UPLOAD_ENDPOINT_URL_TEMPLATE: ClassVar[
        str
    ] = "https://wow.curseforge.com/api/projects/{project_id}/upload-file"

    def upload(
        self,
        *,
        project_id: str,
        file: BinaryIO,
        file_name: str,
        display_name: str,
        changelog: Changelog,
        game_version_ids: Sequence[GameVersionId],
        release_type: str,
    ) -> int:
        """
        Uploads an addon file to Curseforge's WoW addon index and returns its file id.

        `display_name` is the name given to the upload and `file_name` is the name of
        file you download.
        """
        # CF wants a multipart/form-data request with two keys-value pairs:
        # - "metadata": to be equal to the JSON-encoded metadata object they. IMO, this
        #               is a little weird because you can already do key-value pairs
        #               with multipart/form-data -- you don't need another encoding
        # - "file": set to the binary data of the zip
        # source: https://support.curseforge.com/en/support/solutions/articles/9000197321-curseforge-upload-api#Project-Upload-File-API # noqa: E501

        # just an interesting note: you can upload new files with the same
        # metadata all day long -- CF just uses file ids to differentiate and presumably
        # will prefer most-recently-uploaded file. (can't upload dupe zips though, they
        # check for that.)
        response = self._CLIENT.post(
            url=self.UPLOAD_ENDPOINT_URL_TEMPLATE.format(project_id=project_id),
            headers={self.TOKEN_HEADER_NAME: self.api_token},
            data={
                "metadata": json.dumps(
                    {
                        "changelog": changelog.text,
                        "changelogType": changelog.type_,
                        "displayName": display_name,
                        "gameVersions": game_version_ids,
                        "releaseType": release_type,
                    }
                )
            },
            files={"file": (file_name, file, "application/zip")},
        )

        _raise_for_status(response, "upload")

        return response.json()["id"]  # type: ignore

    def get_version_map(self) -> Mapping[str, GameVersionId]:
        response = self._CLIENT.get(
            self.VERSION_ENDPOINT_URL, headers={self.TOKEN_HEADER_NAME: self.api_token}
        )
        _raise_for_status(response, "game version lookup")

        version_map: dict[str, GameVersionId] = {}
        for version_obj in response.json():
            version, id_ = version_obj["name"], version_obj["id"]
            if version not in version_map or version_map[version] < id_:
                version_map[version] = id_

        return version_map

    @classmethod
    def uploaded_file_url(cls, slug: str, file_id: int) -> str:
        return cls.UPLOADED_FILE_URL_TEMPLATE.format(
            slug=slug,
            file_id=file_id,
        )


@frozen(kw_only=True)
class Changelog:
    text: str
    type_: ChangelogType

    CHANGELOG_SUFFIX_MAP: ClassVar[Mapping[str, ChangelogType]] = {
        ".md": "markdown",
        ".markdown": "markdown",
        ".html": "html",
        ".txt": "text",
    }
    DEFAULT_CHANGELOG_TYPE: ClassVar[ChangelogType] = "text"

    @classmethod
    def from_text(cls, text: str, type_: ChangelogType | None = None) -> Changelog:
        if type_ is None:
            type_ = "text"
        return cls(type_=type_, text=text)

    @classmethod
    def from_path(cls, path: Path, type_: ChangelogType | None = None) -> Changelog:
        if type_ is None:
            suffix_normalized = path.suffix.lower()
            if suffix_normalized in cls.CHANGELOG_SUFFIX_MAP:
                type_ = cls.CHANGELOG_SUFFIX_MAP[suffix_normalized]
            else:
                warn(
                    f"Unable to determine changelog type from extension for {path}, "
                    f"so assuming {cls.DEFAULT_CHANGELOG_TYPE}"
                )
                type_ = cls.DEFAULT_CHANGELOG_TYPE

        try:
            contents = path.read_text(encoding="utf-8")
        except UnicodeDecodeError as unicode_decode_error:
            raise EncodingError(
                f'Changelog file "{path}" should be utf-8: {unicode_decode_error}'
            ) from unicode_decode_error
        except FileNotFoundError as file_not_found_error:
            raise PathMissingError(
                f"Changelog path {path} should exist. Please update the path and try "
                "again."
            ) from file_not_found_error

        return cls(
            type_=type_,
            text=contents,
        )

    @classmethod
    def suggest_changelog_type(cls, suffix: str) -> ChangelogType | None:
        return cls.CHANGELOG_SUFFIX_MAP.get(suffix.lower(), None)
